/*! \file tfrfme.h
 *
 * \brief Representation of the trapping calculation objects.
 */

#ifndef INCLUDE_TFRFME_H_
#define INCLUDE_TFRFME_H_

/*! \brief Class to represent the first group of trapping swap data.
 */
class Swap1 {
protected:
  //! NLMMT = 2 * LM * (LM + 2)
  int nlmmt;

  //! QUESTION: definition?
  std::complex<double> *wk;

  /*! \brief Load a Swap1 instance from a HDF5 binary file.
   *
   * \param file_name: `string` Name of the file to be loaded.
   * \return instance: `Swap1 *` Pointer to a new Swap1 instance.
   */
  static Swap1 *from_hdf5(std::string file_name);

  /*! \brief Load a Swap1 instance from a legacy binary file.
   *
   * \param file_name: `string` Name of the file to be loaded.
   * \return instance: `Swap1 *` Pointer to a new Swap1 instance.
   */
  static Swap1 *from_legacy(std::string file_name);

  /*! \brief Save a Swap1 instance to a HDF5 binary file.
   *
   * \param file_name: `string` Name of the file to be written.
   */
  void write_hdf5(std::string file_name);

  /*! \brief Save a Swap1 instance to a legacy binary file.
   *
   * \param file_name: `string` Name of the file to be written.
   */
  void write_legacy(std::string file_name);

public:
  /*! \brief Swap1 instance constructor.
   *
   * \param lm: `int` Maximum field expansion order.
   */
  Swap1(int lm);

  /*! \brief Swap1 instance destroyer.
   */
  ~Swap1() { delete[] wk; }

  /*! \brief Load a Swap1 instance from binary file.
   *
   * \param file_name: `string` Name of the file.
   * \param mode: `string` Format of the file (can be either "HDF5"
   * or "LGEACY". Default is "LEGACY").
   * \return instance: `Swap1 *` Pointer to a newly created Swap1 instance.
   */
  static Swap1* from_binary(std::string file_name, std::string mode="LEGACY");

  /*! \brief Get an element from the WK vector.
   *
   * \param index: `int` Index of the desired element.
   * \return value: `complex<double>` The value of the requested element.
   */
  std::complex<double> get_element(int index) { return wk[index]; }

  /*! \brief Set an element in the WK vector.
   *
   * \param index: `int` Index of the desired element.
   * \param value: `complex<double>` The value to be stored in the vector.
   */
  void set_element(int index, std::complex<double> value) { wk[index] = value; }
  
  /*! \brief Write a Swap1 instance to binary file.
   *
   * \param file_name: `string` Name of the file.
   * \param mode: `string` Format of the file (can be either "HDF5"
   * or "LGEACY". Default is "LEGACY").
   */
  void write_binary(std::string file_name, std::string mode="LEGACY");

  /*! \brief Test whether two instances of Swap1 are equal.
   *
   * \param other: `Swap1 &` Reference to the instance to be compared
   * with.
   * \return result: `bool` True, if the two instances are equal,
   * false otherwise.
   */
  bool operator ==(Swap1 &other);
};

/*! \brief Class to represent the second group of trapping swap data.
 */
class Swap2 {
protected:
  //! Number of vector coordinates
  int nkv;
  //! QUESTION: definition?
  double *vkv;
  //! QUESTION: definition?
  double **vkzm;
  //! QUESTION: definition?
  double apfafa;
  //! QUESTION: definition?
  double pmf;
  //! QUESTION: definition?
  double spd;
  //! QUESTION: definition?
  double rir;
  //! QUESTION: definition?
  double ftcn;
  //! QUESTION: definition?
  double fshmx;
  //! QUESTION: definition?
  double vxyzmx;
  //! Cartesian displacement. QUESTION: correct?
  double delxyz;
  //! QUESTION: definition?
  double vknmx;
  //! QUESTION: definition?
  double delk;
  //! QUESTION: definition?
  double delks;
  //! NLMMT = LM * (LM + 2) * 2
  int nlmmt;
  //! Number of radial vector coordinates. QUESTION: correct?
  int nrvc;

  /*! \brief Load a Swap2 instance from a HDF5 binary file.
   *
   * \param file_name: `string` Name of the file to be loaded.
   * \return instance: `Swap2 *` Pointer to a new Swap2 instance.
   */
  static Swap2 *from_hdf5(std::string file_name);

  /*! \brief Load a Swap2 instance from a legacy binary file.
   *
   * \param file_name: `string` Name of the file to be loaded.
   * \return instance: `Swap2 *` Pointer to a new Swap2 instance.
   */
  static Swap2 *from_legacy(std::string file_name);

  /*! \brief Save a Swap2 instance to a HDF5 binary file.
   *
   * \param file_name: `string` Name of the file to be written.
   */
  void write_hdf5(std::string file_name);

  /*! \brief Save a Swap2 instance to a legacy binary file.
   *
   * \param file_name: `string` Name of the file to be written.
   */
  void write_legacy(std::string file_name);

public:
  /*! \brief Swap2 instance constructor.
   *
   * \param _nkv: `int` Number of vector coordinates. QUESTION: correct?
   */
  Swap2(int _nkv);

  /*! \brief Swap2 instance destroyer.
   */
  ~Swap2();

  /*! \brief Load a Swap2 instance from binary file.
   *
   * \param file_name: `string` Name of the file.
   * \param mode: `string` Format of the file (can be either "HDF5"
   * or "LGEACY". Default is "LEGACY").
   * \return instance: `Swap2 *` Pointer to a newly created Swap2 instance.
   */
  static Swap2* from_binary(std::string file_name, std::string mode="LEGACY");

  /*! \brief Get an element from the VKZM matrix.
   *
   * \param row: `int` Index of the desired element row.
   * \param col: `int` Index of the desired element column.
   * \return value: `double` The value of the requested element.
   */
  double get_matrix_element(int row, int col) { return vkzm[row][col]; }

  /*! \brief Get a parameter by its name.
   *
   * \param param_name: `string` Name of the parameter.
   * \return value: `double` The value of the requested parameter.
   */
  double get_param(std::string param_name);

  /*! \brief Get an element from the VKV vector.
   *
   * \param index: `int` Index of the desired element.
   * \return value: `double` The value of the requested element.
   */
  double get_vector_element(int index) { return vkv[index]; }

  /*! \brief Set an element in the VKZM matrix.
   *
   * \param row: `int` Index of the desired element row.
   * \param col: `int` Index of the desired element column.
   * \param value: `double` The value of the element to be stored.
   */
  void set_matrix_element(int row, int col, double value) { vkzm[row][col] = value; }
  
  /*! \brief Set a parameter by its name and value.
   *
   * \param param_name: `string` Name of the parameter.
   * \param value: `double` The value of the parameter.
   */
  void set_param(std::string param_name, double value);

  /*! \brief Get an element from the VKV vector.
   *
   * \param index: `int` Index of the desired element.
   * \param value: `double` The value of element to be stored.
   */
  void set_vector_element(int index, double value) { vkv[index] = value; }

  /*! \brief Write a Swap2 instance to binary file.
   *
   * \param file_name: `string` Name of the file.
   * \param mode: `string` Format of the file (can be either "HDF5"
   * or "LGEACY". Default is "LEGACY").
   */
  void write_binary(std::string file_name, std::string mode="LEGACY");

  /*! \brief Test whether two instances of Swap2 are equal.
   *
   * \param other: `Swap1 &` Reference to the instance to be compared
   * with.
   * \return result: `bool` True, if the two instances are equal,
   * false otherwise.
   */
  bool operator ==(Swap2 &other);
};

/*! \brief Class to represent the trapping configuration.
 */
class TFRFME {
protected:
  //! NLMMT = 2 * LM * (LM + 2)
  int nlmmt;
  //! NRVC = NXV * NYV * NZV
  int nrvc;
  
  //! Field expansion mode identifier.
  int lmode;
  //! Maximim field expansion order;
  int lm;
  //! QUESTION: definition?
  int nkv;
  //! Number of computed X coordinates.
  int nxv;
  //! Number of computed Y coordinates.
  int nyv;
  //! Number of computed Z coordinates.
  int nzv;
  //! Wave number in scale units
  double vk;
  //! External medium refractive index
  double exri;
  //! QUESTION: definition?
  double an;
  //! QUESTION: definition?
  double ff;
  //! QUESTION: definition?
  double tra;
  //! QUESTION: definition?
  double spd;
  //! QUESTION: definition?
  double frsh;
  //! QUESTION: definition?
  double exril;
  //! Vector of computed x positions
  double *xv;
  //! Vector of computed y positions
  double *yv;
  //! Vector of computed z positions
  double *zv;
  //! QUESTION: definition?
  std::complex<double> **wsum;

  /*! \brief Load a configuration instance from a HDF5 binary file.
   *
   * \param file_name: `string` Name of the file to be loaded.
   * \return instance: `TFRFME *` Pointer to a new trapping configuration
   * instance.
   */
  static TFRFME *from_hdf5(std::string file_name);

  /*! \brief Load a configuration instance from a legacy binary file.
   *
   * \param file_name: `string` Name of the file to be loaded.
   * \return instance: `TFRFME *` Pointer to a new trapping configuration
   * instance.
   */
  static TFRFME *from_legacy(std::string file_name);

  /*! \brief Save a configuration instance to a HDF5 binary file.
   *
   * \param file_name: `string` Name of the file to be written.
   */
  void write_hdf5(std::string file_name);

  /*! \brief Save a configuration instance to a legacy binary file.
   *
   * \param file_name: `string` Name of the file to be written.
   */
  void write_legacy(std::string file_name);

public:
  /*! \brief Trapping configuration instance constructor.
   *
   * \param _lmode: `int` Order expansion mode flag.
   * \param _lm: `int` Maximum field expansion order.
   * \param _nkv: `int` Number of wave vector coordinates. QUESTION: correct?
   * \param _nxv: `int` Number of computed X coordinates.
   * \param _nyv: `int` Number of computed Y coordinates.
   * \param _nzv: `int` Number of computed Z coordinates.
   */
  TFRFME(
	 int _lmode, int _lm, int _nkv, int _nxv, int _nyv, int _nzv
);
  
  /*! \brief Trapping configuration instance destroyer.
   */
  ~TFRFME();

  /*! \brief Load a trapping configuration instance from binary file.
   *
   * \param file_name: `string` Name of the file.
   * \param mode: `string` Format of the file (can be either "HDF5"
   * or "LGEACY". Default is "LEGACY").
   * \return instance: `TFRFME *` Pointer to a newly created configuration
   * instance.
   */
  static TFRFME* from_binary(std::string file_name, std::string mode="LEGACY");

  /*! \brief Get an element from the WSUM matrix.
   *
   * \param row: `int` Row index of the element to be read.
   * \param col: `int` Column index of the element to be read.
   * \return value: `complex<double>` The value of the requested element.
   */
  std::complex<double> get_matrix_element(int row, int col);

  /*! \brief Get a configuration parameter.
   *
   * \param param_name: `string` Name of the parameter.
   * \return value: `double` The value of the requested parameter.
   */
  double get_param(std::string param_name);

  /*! \brief Get the X coordinate of the computed point at given index.
   *
   * \param index: `int` Index of the requested point.
   * \return x: `double` X coordinate of the requested point.
   */
  double get_x(int index) { return xv[index]; }

  /*! \brief Get the Y coordinate of the computed point at given index.
   *
   * \param index: `int` Index of the requested point.
   * \return y: `double` Y coordinate of the requested point.
   */
  double get_y(int index) { return yv[index]; }

  /*! \brief Get the Z coordinate of the computed point at given index.
   *
   * \param index: `int` Index of the requested point.
   * \return z: `double` Z coordinate of the requested point.
   */
  double get_z(int index) { return zv[index]; }

  /*! \brief Set an element in the WSUM matrix.
   *
   * \param row: `int` Row index of the element to be read.
   * \param col: `int` Column index of the element to be read.
   * \param value: `complex<double>` The value to be placed in the matrix.
   */
  void set_matrix_element(int row, int col, std::complex<double> value);
  
  /*! \brief Set a configuration parameter.
   *
   * \param param_name: `string` Name of the parameter.
   * \param value: `double` Value to be stored as parameter.
   */
  void set_param(std::string param_name, double value);

  /*! \brief Set the X coordinate of the computed point at given index.
   *
   * \param index: `int` Index of the requested point.
   * \param value: `double` X coordinate of the requested point.
   */
  void set_x(int index, double value) { xv[index] = value; }

  /*! \brief Set the Y coordinate of the computed point at given index.
   *
   * \param index: `int` Index of the requested point.
   * \param value: `double` Y coordinate of the requested point.
   */
  void set_y(int index, double value) { yv[index] = value; }

  /*! \brief Set the Z coordinate of the computed point at given index.
   *
   * \param index: `int` Index of the requested point.
   * \param value: `double` Z coordinate of the requested point.
   */
  void set_z(int index, double value) { zv[index] = value; }

  /*! \brief Write a trapping configuration instance to binary file.
   *
   * \param file_name: `string` Name of the file.
   * \param mode: `string` Format of the file (can be either "HDF5"
   * or "LGEACY". Default is "LEGACY").
   */
  void write_binary(std::string file_name, std::string mode="LEGACY");
  
  /*! \brief Test whether two instances of configuration are equal.
   *
   * \param other: `TFRFME &` Reference to the instance to be compared
   * with.
   * \return result: `bool` True, if the two instances are equal,
   * false otherwise.
   */
  bool operator ==(TFRFME &other);
};
#endif
